program WEIcheckpar
  ! Reads parameters for WEI migration code that uses MPI, except nws and nzs.
  ! Reads some parameters specific only to this program, too.
  ! Makes some elementary checks, then suggests the nws and nzs values so that
  ! the nodes will communicate with the master as few times as possible

  use sep, only: sep_init, from_param, sep_close
  implicit none

  ! Input parameters that are NOT found in WEI parameter files:
  real    :: ram    ! RAM/node, in Gigabytes. Default 2.
  integer :: nnodes ! Number of nodes.
  integer :: proc   ! Processors per node. Default 2.
  integer :: f      ! Max RAM percentage to be used by a data slice. Default 25
  integer :: esize  ! Esize. Default 8.
  ! Input parameters that ARE found in WEI parameter files:
  integer :: aw__n  ! Total number of frequencies in the data.
  integer :: az__n  ! Total number of depth steps in the data.
  integer :: amx_n  ! Total number of midpoints in the data in the x direction
  integer :: amy_n  ! Total number of midpoints in the data in the y direction
  integer :: ahx_n  ! Total number of offsets   in the data in the x direction
  integer :: ahy_n  ! Total number of offsets   in the data in the y direction

  ! Outputs (on the screen, not in a file) - one or several suggestions for
  integer :: nws    ! Number of frequencies in a data slice
  integer :: nzs    ! Number of depth steps in a data slice

  ! Other variables in the code:
  integer :: maxblk ! max # of elementary blocks that will fit into memory
  integer :: iw,iz  ! Counting indices
  integer, dimension(:), allocatable:: nws_list ! Possible values for nws
  integer, dimension(:), allocatable:: nzs_list ! Possible values for nzs

  call sep_init
  call from_param('ram'   , ram   ,  2.)
  call from_param('nnodes', nnodes     )
  call from_param('proc'  , proc  ,  2 )
  call from_param('f'     , f     , 25 )
  call from_param('esize' , esize ,  8 )
  call from_param('aw__n' , aw__n      )
  call from_param('az__n' , az__n      )
  call from_param('amx_n' , amx_n      )
  call from_param('amy_n' , amy_n      )
  call from_param('ahx_n' , ahx_n      )
  call from_param('ahy_n' , ahy_n      )
  call sep_close

  ! Elementary verifications

  if (ram<=0) call seperr('ram cannot be negative')
  if (ram>20) call seperr('Specify ram in Gb, this is too big to be in Gb')
  ram = ram * 2**30 ! Convert from Gb to bytes

  if (nnodes<  1) call seperr('nnodes too small')
  if (nnodes>500) call seperr('nnodes unbelievably big')

  if (proc< 1) call seperr('proc too small')
  if (proc>16) call seperr('proc ')

  if (f<=  0) call seperr('f should be greater than zero')
  if (f>=100) call seperr('f should be smaller than 100')

  if (esize/=4 .and. esize/=8) call seperr('esize should be 4 or 8')

  if (aw__n < 1) call seperr('aw__n  too small')
  if (az__n < 1) call seperr('az__n  too small')
  if (amx_n < 1) call seperr('amx_n too small')
  if (amy_n < 1) call seperr('amy_n too small')
  if (ahx_n < 1) call seperr('ahx_n too small')
  if (ahy_n < 1) call seperr('ahy_n too small')
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

  ! Total # of processors must divide total # of frequencies
  if (mod(aw__n,nnodes*proc)/=0) call seperr('nnodes*proc must be a factor of aw__n')
  allocate(nws_list(aw__n), nzs_list(az__n))
  nws_list = (/1:aw__n/)
  nzs_list = (/1:az__n/)

  ! nzs must divide az__n
  where(mod(az__n,nzs_list)/=0) nzs_list=0

  ! proc must divide nws
  where(mod(nws_list,proc)/=0) nws_list=0

  ! nws must divide aw__n
  where(nws_list/=0.and.mod(aw__n,nws_list)/=0) nws_list=0

  ! Find max number of elementary blocks that will fit into memory
  maxblk = ram * real(f) / real(amx_n * amy_n * ahx_n * ahy_n * esize)

  ! Find nzs and nws
  nzs = 0
  nws = 0

  z_loop: do iz = az__n, 1, -1
     z_if: if (nzs_list(iz)/=0) then
        w_loop: do iw = aw__n, 1, -1
           w_if: if(nws_list(iw)/=0) then
              maxsize_if: if (nzs_list(iz) * nws_list(iw) <= maxblk) then
                 write(0,*) maxblk-nzs_list(iz) * nws_list(iw),nzs_list(iz), nws_list(iw)
                 fillram_if: if (nzs_list(iz) * nws_list(iw) > nzs * nws) then
                    nzs = nzs_list(iz)
                    nws = nws_list(iw)
                 end if fillram_if
              end if maxsize_if
           end if w_if
        end do w_loop
     end if z_if
  end do z_loop

  if (nzs*nws==0) then
     write(0,*) 'Elementary block (amx_n * amy_n * ahx_n * ahy_n * esize) too'
     write(0,*) 'big to fit in RAM. Increase f or window data'
  else
     write(0,*) '@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@'
     write(0,*) '@@ nws=',nws, '@@@@@@@@ nzs=',nzs,'@@'
     write(0,*) '@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@'
  end if

  deallocate(nws_list, nzs_list)
end program WEIcheckpar
